unit Llistint;

{ LListint unit: doubly linked integer list-component 02.Jan.1997 }
{ by Alexander Wernhart, wea@felten.co.at }
{ Version History:                                                }
{ V1.0:                                                           }
{   02.Jan.1997 - initial revision                                }
{ V1.1:                                                           }
{   15.Feb.1997 - method clear added                              }

{ based on: }

{ TLink unit: doubly linked lists 5/22/95}
{ by Jeff Atwood, JAtwood159@AOL.COM. }
{ }
{ This unit can be used for stacks, deques, and free lists too.  }
{ }
{ I couldn't find a doubly-linked list implemented as an object ANYWHERE }
{ so I wrote it myself, after much trial, error, and poring over         }
{ obscure programming reference books. Hey-- it's not brain surgery, but }
{ pointers can be so naughty. }
{ }
{ For simplicity's sake, and to keep this a one-day project, I am only    }
{ storing simple integers in the cells. You can easily, easily change     }
{ that to any data type supported by Delphi including records. I would    }
{ NOT recommend trying to store a whole object with methods in there...   }
{ I couldn't get that to work. But if you can, E-Mail me. I don't know if }
{ it's even possible. }
{ }
{ There is one main object, which uses the "CELL" record type for each    }
{ entry in the list. I don't know how to hide the CELL record type from   }
{ the user, but it should be internal to this unit. The main object is    }
{ the TLink, which keeps track of the size, first, last, and current      }
{ cell records. You can move around in the list by using the Move methods }
{ and find using the Seek method. It's all fairly straightforward, look   }
{ at the demo form for examples, there are also comments in the code.     }
{ }
{ If you're feeling ambitious, I recommend you modify the cell record to  }
{ store pointers instead of integers. Don't forget to make copies of the  }
{ data, because if you point to the actual location, you're screwed when  }
{ the user destroys that instance. You gotta copy it... How many times    }
{ did I get burned by THAT one?? Also, it would be cool to turn this into }
{ a VCL component, if anyone wants to do that. }
{ }
{ This code is freeware. Please E-Mail me any cool additions, bug fixes,  }
{ rants, raves, etc. at JAtwood159@AOL.COM! Thanks for trying my code, I  }
{ hope it helps someone... }

interface

uses
  SysUtils, WinTypes, WinProcs, Messages, Classes, Graphics, Controls,
  Forms, Dialogs;

type
  TData = integer;  { change, if you want to store other data type }

  CellPtr = ^TCell;

  TCell = record
    data: TData;
    next: CellPtr;
    prev: CellPtr;
  end;

  TLListInt = class(TComponent)
  private
    { Private-Deklarationen }
    top: CellPtr;
    bottom: CellPtr;
    current: CellPtr;
    FCurrPos: integer;
    FSize: integer;
    function SetItemIndex(Position: Integer): Boolean;
    function GetItem(Index: Integer): TData;
    procedure SetItem(Index: Integer; newData:TData);
    procedure DelAll;
  protected
    { Protected-Deklarationen }
  public
    { Public-Deklarationen }
    constructor create(AOwner: TComponent); override;
    destructor destroy; override;
    procedure Add(newItem: TData);
    procedure clear;
    function Insert(newItem: TData; Index: integer): Boolean;
    function Delete(Index: integer): Boolean;
    function FindFirst(cmpItem: TData): integer;
    function FindNext(cmpItem: TData): integer;
    property ItemCount: integer read FSize;
    property ItemIndex: integer read FCurrPos write SetItemIndex;
    property Item[Index: Integer]: TData read GetItem write SetItem;
  published
    { Published-Deklarationen }
  end;

procedure Register;

implementation

{ set up the TLListInt object with default values }
constructor TLListInt.create(AOwner: TComponent);
begin
  inherited create(AOwner);
  top := nil;
  bottom := nil;
  current := nil;
  FCurrPos := -1;
  FSize := 0;
end;

{ destroy the entire list, cell by cell }
destructor TLListInt.destroy;
begin
  DelAll;
  inherited destroy;
end;

{ destroy the entire list, cell by cell }
procedure TLListInt.clear;
begin
  DelAll;
end;

{ destroy the entire list, cell by cell }
procedure TLListInt.DelAll;
var
  curCell: CellPtr;
  nextCell: CellPtr;
begin
  curCell := top;
  while not (curCell = nil) do begin
    nextCell := curCell^.next;
    freemem(curCell, SizeOf(TCell));
    curCell := nextCell;
  end;
  top := nil;
  bottom := nil;
  current := nil;
  FCurrPos := -1;
  FSize := 0;
end;

{ make current the absolute cell at Position in the list }
{ where top = 0 }
function TLListInt.SetItemIndex(Position: Integer): Boolean;
var
  curCell: CellPtr;
  cnt: longint;
  SeekTop: Boolean;
begin
  result := false;
  SeekTop := false;
  { see if Position is valid }
  if (Position < 0) or (Position >= Fsize) then
  begin
    messagedlg('Index out of Range',mtError,[mbOK],0);
  end else
  begin
    { Did we set a position before? }
    if FCurrPos > -1 then
    begin
      { optimize search, if user browses through list(forward or backward) }
      { or just reads an item a second time }
      case ((Position - FCurrPos) + 1) of
        { new item is located before current item }
        0: begin
             current := current^.prev;
             FCurrPos := FCurrPos - 1;
             result := true;
           end;
        { new item is current item }
        1: result := true;
        { new item is located after current item }
        2: begin
             current := current^.next;
             FCurrPos := FCurrPos + 1;
             result := true;
           end;
        { new item is located somewhere else in the list, let's seek from top }
        else
        begin
          SeekTop := true;
        end;
      end;
    { we have no current item, let's seek from top }
    end else
    begin
      SeekTop := true;
    end;
  end;
  { seek from the top of the list cell by cell }
  if seekTop then
  begin
    curCell := top;
    cnt := 0;
    while (curCell <> nil) do
    begin
      if cnt = Position then
      begin
        FCurrPos := cnt;
        current := CurCell;
        result := true;
        curCell := nil;
      end else
      begin
        curCell := curCell^.next;
        cnt := cnt + 1;
      end;
    end;
  end;
end;


{ insert cell at bottom of list }
procedure TLListInt.Add(newItem: TData);
var
  newCell: CellPtr;
begin
  GetMem(newCell, Sizeof(TCell));
  newCell^.data := newItem;
  newCell^.prev := bottom;
  newCell^.next := nil;
  { special case: this is first cell added }
  if bottom = nil then
    top := newCell
  else
    bottom^.next := newCell;
  bottom := newCell;
  FSize := FSize + 1;
end;

{ insert cell before item[Index] }
function TLListInt.Insert(newItem: TData; Index: integer): Boolean;
var
  newCell: CellPtr;
begin
  result := false;
  if SetItemIndex(Index) then
  begin
    GetMem(newCell, Sizeof(TCell));
    newCell^.data := newItem;
    newCell^.prev := current^.prev;
    newCell^.next := current;
    { special case: current cell is first cell }
    if current^.prev = nil then
      top := newCell
    else
      current^.prev^.next := newCell;
    current^.prev := newCell;
    FSize := FSize + 1;
    FCurrPos := FCurrPos + 1;
    result := True;
  end;
end;

{ delete the current cell }
function TLListInt.Delete(Index: integer): Boolean;
var
  noDelete: Boolean;
begin
  { we can only delete the current record }
  if SetItemIndex(Index) then
  begin
    noDelete := true;
    { see if list has one item }
    if FSize = 1 then
    begin
      top := nil;
      bottom := nil;
      noDelete := false;
    end;
    { see if we're at the top of list }
    if (current^.prev = nil) and noDelete then
    begin
      top := current^.next;
      top^.prev := nil;
      noDelete := false;
    end;
    { see if we're at the bottom of list }
    if (current^.next = nil) and noDelete then
    begin
      bottom := current^.prev;
      bottom^.next := nil;
      noDelete := false;
    end;
    { we must be in middle of list of size > 1 }
    if noDelete then
    begin
      current^.prev^.next := current^.next;
      current^.next^.prev := current^.prev;
      noDelete := false;
    end;

    if not noDelete then
    begin
      result := True;
      freemem(current, SizeOf(TCell));
      current := nil;
      FCurrPos := -1;
      FSize := FSize - 1;
      if FSize = 0 then
      begin
        top := nil;
        bottom := nil;
      end;
    end;
  end;
end;

{ make current the absolute cell N in the list }
{ where top = 0 }
{ return data item from current list position }
function TLListInt.GetItem(Index: Integer): TData;
begin
  if SetItemIndex(Index) then
  begin
    result := current^.data;
  end;
end;

{ make current the absolute cell N in the list }
{ where top = 0 }
{ set data item to current list position }
procedure TLListInt.SetItem(Index: Integer; newData: TData);
begin
  if SetItemIndex(Index) then
  begin
    current^.data := newData;
  end;
end;

{ Look for item in data field. Starts at top of list }
{ and looks at every item until a match is found.    }
{ if found, makes matched cell current, and returns  }
{ absolute location of match where 0 = top.          }
{ if not found returns -1                            }
function TLListInt.FindFirst(cmpItem: TData): integer;
var
  curCell: CellPtr;
  cnt: longInt;
begin
  result := -1;
  curCell := top;
  cnt := 0;
  while (curCell <> nil) do
  begin
    if curCell^.Data = cmpItem then
    begin
      current := curCell;
      FCurrPos := cnt;
      result := cnt;
      curCell := nil;
    end else
    begin
      curCell := curCell^.next;
      cnt := cnt + 1;
    end;
  end;
end;

{ Look for item in data field. Starts at current^.next}
{ and looks at every item until a match is found.    }
{ if found, makes matched cell current, and returns  }
{ absolute location of match where 0 = top.          }
{ if not found, returns -1                           }
function TLListInt.FindNext(cmpItem: TData): integer;
var
  curCell: CellPtr;
  cnt: longInt;
begin
  result := -1;
  curCell := current^.next;
  cnt := FCurrPos + 1;
  while (curCell <> nil) do
  begin
    if curCell^.Data = cmpItem then
    begin
      current := curCell;
      FCurrPos := cnt;
      result := cnt;
      curCell := nil;
    end else
    begin
      curCell := curCell^.next;
      cnt := cnt + 1;
    end;
  end;
end;

procedure Register;
begin
  RegisterComponents('Samples', [TLListInt]);
end;

end.
